;######################################################################
;
; Copyright (C) 2004-2017, Michele Cappellari
; E-mail: cappellari_at_astro.ox.ac.uk
;
; This software is provided as is without any warranty whatsoever.
; Permission to use, for non-commercial purposes is granted.
; Permission to modify for personal or internal use is granted,
; provided this copyright and disclaimer are included unchanged
; at the beginning of the file. All other rights are reserved.
;
;######################################################################
pro cap_symm_rotate_points, x, y, ang, xNew, yNew
compile_opt idl2, hidden
;
; Michele cappellari, Leiden, 15 August 2003

theta = (ang - 90.0)/!RADEG  ; Measure from positive Y axis
xNew = x*COS(theta) - y*SIN(theta)
yNew = x*SIN(theta) + y*COS(theta)

end
;----------------------------------------------------------------------
pro cap_symmetrize_velfield, xbin, ybin, velBin, velSym, SYM=sym, PA=pa, TRIANG=tri
compile_opt idl2
;
; This routine generates a bi-symmetric ('axisymmetric')
; version of a given set of kinematical measurements.
; PA: is the angle in degrees, measured counter-clockwise,
;   from the vertical axis (Y axis) to the galaxy major axis.
; SYM: by-simmetry: is 1 for (V,h3,h5) and 2 for (sigma,h4,h6)
;
; HISTORY:
;
; V1.0.0: Michele Cappellari, Vicenza, 21 May 2004
; V1.0.1: Added MISSING keyword to TRIGRID call. Flipped velocity sign.
;   Written basic documentation. MC, Leiden, 25 May 2004
; V1.1.0: Included point-symmetric case. Remco van den Bosch, Leiden, 18 January 2005
; V1.1.1: Minor code revisions. MC, Leiden, 23 May 2005
; V1.1.2: Important: changed definition of PA to be measured counterclockwise
;   with respect to the positive Y axis, as in astronomical convention and
;   consistently with my FIND_GALAXY routine. MC, Leiden, 1 June 2005
; V1.1.3: Added optional keyword TRIANG. Corrected rare situation with w=-1.
;   MC, Leiden, 2 June 2005
; V1.1.4: Added prefix SYMM_ to internal functions to prevent conflicts
;   with external functions with the same name. MC, Oxford, 11 May 2007
; V2.0.0 : Completely rewritten without any loop. MC, Oxford, 8 October 2013
; V2.0.1: Uses TOLERANCE keyword of TRIANGULATE to try to avoid IDL error
;   "TRIANGULATE: Points are co-linear, no solution." MC, Oxford, 2 December 2013
; V2.0.2: Made compatible with IDL < 8. MC, Oxford, 2 February 2017
;
on_error, 2

n = n_elements(xbin)
if n_elements(pa) eq 0 then pa = 0.0
cap_symm_rotate_points, xbin, ybin, -pa, x, y  ; Negative PA for counter-clockwise
if ~n_elements(tri) then $
    TRIANGULATE, x, y, tri, TOLERANCE=1d-12*max(abs([x, y]))

xout = [x,-x, x,-x]
yout = [y, y,-y,-y]
velOut = griddata(x, y, velBin, XOUT=xout, YOUT=yout, $
    MISSING='NaN', /LINEAR, TRIANGLES=tri)
velOut = reform(velOut, n, 4)
if sym eq 1 then velOut[*, [1, 3]] *= -1.0
if !Version.Release GE 8 then $
    velSym = mean(velOut, DIMENSION=2, /NAN) $
else begin
    velSym = dblarr(n, /NOZERO)
    for j=0,n-1 do $
        velSym[j] = mean(velOut[j, *], /NAN)
endelse

end
;----------------------------------------------------------------------
